/* hdb.h
** hash database on disk
** wcm, 2010.05.27 - 2010.10.28
** ===
*/

#ifndef HDB_H
#define HDB_H 1

/* libc: */
#include <stddef.h>  /* size_t */
#include <stdint.h>  /* uint32_t */

/* unix: */
#include <unistd.h>  /* off_t */

/* lasagna: */
#include "ioq.h"
#include "dynbuf.h"
#include "tain.h"
#include "uchar.h"

/*
** hdb32 file structure:
** 
**   I  M  F  P  R  H
** 
** where:
** 
**   I: identifier section (32 bytes)
**       16 bytes: identifier "hdb32/1.0\0\0\0\0\0\0\0"
**        4 bytes: pid of creator process
**       12 bytes: TAI64N timestamp
** 
**   M: metadata section (8 bytes)
**       4 bytes ntabs: number of hash tables (configured)
**       4 bytes nrecs: total records in file (recorded)
**
**   F: hash function parameter section (8 bytes)
**       4 bytes hf_init: hash function initial value
**       4 bytes hf_mult: hash function multiplier
** 
**   P: hash table pointer section (8 bytes * ntabs)
**       P[0..(ntabs-1)] hash subtable pointers
**       ntabs hash subtable pointers to ntabs subtables
**       each pointer comprised of:
**           4 bytes tbase: offset to base of subtable in section H
**           4 bytes slots: number of slots in subtable
** 
**   R: record data section
**       R[0..(nrecs-1)] records
**       nrecs records, each record comprised of:
**           4 bytes klen: length of key in this record
**           4 bytes dlen: length of data in this record
**           klen bytes: key
**           dlen bytes: data
** 
**   H: hash table section
**       H[0..(ntabs-1)][0..(slots[t])] hash subtables
**       ntabs subtables of slots[t] slots per subtable
**       each slot comprised of:
**           4 bytes: hash of key for record inserted in this slot [*1]
**           4 bytes: rpos, offset to record entry or 0 (empty)
**
**   Notes:
**       1.  The key hash value included in each hash table slot is not
**           strictly required to resolve record searches, but improves
**           search performance in cases of key collisions (enables successive
**           entries in the hash table to be scanned directly, rather than
**           performing a separate record lookup for each collison test).
**
**       2.  Generator pid/timestamp may be used to support hash seed
**           initialization in future implementations.
*/


/*
** --- hdb ---
*/

/* file header string (10 bytes): */
#define HDB_IDENT  "hdb32/1.0\0\0\0\0\0\0\0"

/*
** struct hdb:  hdb32 reader object
*/
struct hdb {
  int       fd;
  /* mmap() on fd: */
  uchar_t  *map;
  uint32_t  map_size;
  /* file header: */
  char      ident[1 + sizeof HDB_IDENT]; /* NUL-terminated buffer for ident string */
  uint32_t  pid;   /* pid of creator process */
  tain_t    tain;  /* tain timestamp at creation */
  /* metadata: */
  uint32_t  ntabs; /* number of hash tables */
  uint32_t  nrecs; /* total number of records */
  /* hash function parameters: */
  uint32_t  hf_init; /* hash function initial value */
  uint32_t  hf_mult; /* hash function multiplier */
  /* internal offsets: */
  uint32_t  pbase; /* base offset to hash table pointers */
  uint32_t  rbase; /* base offset to record data */
  uint32_t  hbase; /* base offset to hash table (also eod) */
  /* find parameters: */
  uchar_t  *key;   /* pointer to key (as supplied by hdb_find()) */
  uint32_t  klen;  /* length of key (as supplied by hdb_find()) */
  /* find internals: */
  uint32_t  h;      /* hash: hdb_hash(key, klen) */
  uint32_t  tbase;  /* base offset of hash subtable */
  uint32_t  tslots; /* slots in hash subtable */
  uint32_t  s0;     /* slot probe begins (initial target slot) */
  uint32_t  sN;     /* slot probed for last record found */
  /* retrieval cursors (to found record data): */
  uint32_t  rpos;  /* offset of found record data */
  uint32_t  dlen;  /* length of found record data */
};
typedef struct hdb  hdb_t;


/* hdb_init()
**   initialize an hdb_t object from open file descriptor fd
**   required before any operation
*)*
**   return:
**     0 : success
**    -1 : error, errno set
**
**   note:
**     pair with hdb_clear() on completion
*/
extern int hdb_init(hdb_t *H, int fd);

/* hdb_clear()
**   clear/reset hdb object when finished and/or for reuse
**
**   return:
**     0 : success
**    -1 : error, errno set
**
**   note:
**     application responsible to close/handle any open fd
**     usually paired with hdb_init()
*/
extern int hdb_clear(hdb_t *H);

/* hdb_open()
**   open fd for path then initialize hdb_t object with hdb_init()
**   return:
**     0 : success
**    -1 : error, errno set
**
**   convenience for:
**       fd = open(path, O_RDONLY | O_NONBLOCK);
**       hdb_init(H, fd); 
**     
**   notes:
**     fd is opened: fd = open(path, O_RDONLY | O_NONBLOCK);
**     usually paired with hdb_close() on completion
*/
extern int hdb_open(hdb_t *H, const char *path);

/* hdb_close()
**   close fd for hdb and reset with hdb_clear()
**   return:
**     0 : success
**    -1 : error, errno set
**
**   note:
**     usually paired with hdb_open()
*/
extern int hdb_close(hdb_t *H);

/* hdb_ident()
**   after hdb_init(), access ident string in file header
*/
#define hdb_ident(H)  (H)->ident

/* hdb_ntabs()
**   after hdb_init(), access number of hash tables in file metadata
*/
#define hdb_ntabs(H)  (H)->ntabs

/* hdb_nrecs()
**   after hdb_init(), access number of records in file metadata
*/
#define hdb_nrecs(H)  (H)->nrecs

/* hdb_pbase()
**   after hdb_init(), find computed offset to start of hash table pointers
*/
#define hdb_pbase(H)  (H)->pbase

/* hdb_rbase()
**   after hdb_init(), find computed offset to start of record block
*/
#define hdb_rbase(H)  (H)->rbase

/* hdb_fd()
**   after hdb_open() (or as set in hdb_init()), get file descriptor
*/
#define hdb_fd(H)  (H)->fd

/* hdb_map()
**   after hdb_init(), access mmap()'d file
**   (valid map if non-NULL)
*/
#define hdb_map(H)  (H)->map

/* hdb_map()
**   after hdb_init(), size of mmap()'d file
**   (valid if map non-NULL)
*/
#define hdb_mapsize(H)  (H)->map_size

/* hdb_find()
**   initialize and begin first search of hdb for key of length klen
**   return:
**     1 : found
**     0 : not found
**    -1 : error, errno set
*/
extern int hdb_find(hdb_t *H, const uchar_t *key, uint32_t klen);

/* hdb_findnext()
**   after successful hdb_find(), continue search for duplicate keys
**   return:
**     1 : found
**     0 : not found
**    -1 : error, errno set
*/
extern int hdb_findnext(hdb_t *H);

/* hdb_dlen()
**   after succesful hdb_find(), return size of associated data
**   note: used to prepare/size buffer for subsequent hdb_get()
**   return: H->dlen
*/
#define  hdb_dlen(H)  ((H)->dlen)

/* hdb_dpos()
**   after succesful hdb_find(), return position of associated data
**   note: used to prepare/size buffer for subsequent hdb_get()
**   return: computed from H->rpos
*/
#define  hdb_dpos(H)  ((H)->rpos + 8 + (H)->klen)

/* hdb_klen()
**   after successful hdb_find(), return size of associated key
**   return: H->klen
*/
#define  hdb_klen(H)  ((H)->klen)

/* hdb_kpos()
**   after successful hdb_find(), return position of associated key
**   return: computed from H->rpos
*/
#define  hdb_kpos(H)  ((H)->rpos + 8)

/* hdb_rpos()
**   after successful hdb_find(), return position of associated record
**   return:  H->rpos
*/
#define  hdb_rpos(H)  ((H)->rpos)

/* hdb_get()
**   after successful hdb_find():
**     copy len bytes of record data at hdb_dpos(H) into buffer
**     buffer supplied by caller is at least len bytes
**     if (len > H->dlen), only dlen bytes copied into buffer 
**
**   return:
**     0 : success
**    -1 : error, errno set 
*/
extern int hdb_get(hdb_t *H, uchar_t *buffer, size_t len);

/* hdb_dynget()
**   after successful hdb_find():
**     copy H->dlen bytes record data at H->dpos cursor into dynbuf
**     copy into dynbuf starts/appends from D->buf[D->p]
**     dynbuf grown as necessary
**
**   return:
**     0 : success
**    -1 : error, errno set
**
**   notes:
**     for copy (not append) into D->buf[0], use dynbuf_CLEAR(D)
**     before each hdb_dynget()
**
*/
extern int hdb_dynget(hdb_t *H, dynbuf *D);

/* hdb_distance()
**   after successful hdb_find(), return "distance" from target hash slot
**   note: may be used as an indicator for hash efficiency
**   return:
**     0 : matching record key found in first probe
**     1 : matching record key found in second probe
**     ...
*/
extern uint32_t hdb_distance(hdb_t *H);


/* hdb_seq0()
**   initiate sequential access to records starting from record 0
**   return:
**     1 : success, positioned to record 0
**     0 : eod (no records in database!)
**    -1 : error, errno set
*/
extern int hdb_seq0(hdb_t *H);

/* hdb_seqnext()
**   sequential access to next record
**   return:
**     1 : success, positioned to next record
**     0 : eod
**    -1 : error, errno set
*/
extern int hdb_seqnext(hdb_t *H);

/* hdb_read()
**   from offset in H, read/copy len bytes into buf
**   a low-level operation to provide arbitrary access into an hdb
**
**   return:
**     0 : success
**    -1 : error, errno set
**
**   notes:
**     * caller responsible for providing buf of at least len
**     * caller responsible for providing "meaningful" offset
*/
extern int hdb_read(hdb_t *H, uchar_t *buf, uint32_t len, uint32_t offset);

/* hdb_dynread()
**   from offset in H, read/copy len bytes into dynbuf
**   a low-level operation to provide arbitrary access into an hdb
**   copy into dynbuf appends from D->buf[D->p]
**   (dynbuf grown as necessary)
**    
**   return:
**     0 : success
**    -1 : error, errno set
**
**   notes:
**     for copy (not append) into D->buf[0], use dynbuf_CLEAR(D)
**     before each hdb_dynread()
*/
extern int hdb_dynread(hdb_t *H, dynbuf *B, uint32_t len, uint32_t offset);


/*
** --- hash function ---
*/

extern
uint32_t
hdb_hash(const uchar_t *key, size_t klen, uint32_t init, uint32_t mult);

/* enable avalanche postmix: */
//#define HDB_HFSTIR  1


/*
** --- key insertion/lookup algorithms ---
*/

/* xor high word into low word: */
#define  hdb_WXOR(u)  ((u) ^ ((u)>>16))
/* swap high/low word: */
#define  hdb_WSWAP(u)  (((u)<<16) | ((u)>>16))
/* rotate low byte right: */
#define  hdb_LBX(u)  (((u)>>8) | ((u)<<24))

/* table/slot algorithms: */
/* basis for partitioning hash value among ntabs subtables: */
#define  hdb_NTAB(hash, ntabs)  ((hash) % (ntabs))
  
/* basis for selection of slot within tslots of subtable: */
//#define  hdb_SLOT(hash, tslots)  (hdb_WXOR((hash)) % (tslots))
//#define  hdb_SLOT(hash, tslots)  ((hash) % (tslots))
//#define  hdb_SLOT(hash, tslots)  (hdb_LBX((hash)) % (tslots))
#define  hdb_SLOT(hash, ntabs, tslots) \
  ( (ntabs > 1) ? \
        (hdb_WXOR((hash)) % (tslots)) : \
        ((hash) % (tslots)) )

/* incremental hashing support provided for hdbmk:
**
**   h = hf_init;
**   while(klen){
**       h = hdb_hashnext(h, *key, mult);
**       ++key; --kley;
**   }
**   h = hdb_hashstir(h);
**
**   //or
**
**   h = hf_init;
**   while(part){
**       h = hdb_hashpart(h, *keypart, partlen, mult);
**       ...
**   }
**   h = hdb_hashstir(h); 
*/
extern
uint32_t
hdb_hashnext(uint32_t h, uchar_t  c, uint32_t mult);

extern
uint32_t
hdb_hashpart(uint32_t h, uchar_t *keypart, size_t keypart_len, uint32_t mult);

/* post hash avalanche mix (enabled ifdef HDB_HFSTIR): */
extern
uint32_t
hdb_hashstir(uint32_t h);


#endif /* HDB_H */
/* eof (hdb.h) */
