/* hdbmk.h
** hdb: hash database on disk
** hdbmk: hdb generator (writer)
** wcm, 2010.05.27 - 2010.06.17
** ===
*/

#ifndef HDBMK_H
#define HDBMK_H 1

/* libc: */
#include <stddef.h>  /* size_t */
#include <stdint.h>  /* uint32_t */

/* unix: */
#include <unistd.h>  /* off_t */

/* lasagna: */
#include "ioq.h"
#include "dynbuf.h"
#include "tain.h"
#include "uchar.h"

#include "hdb.h"


/*
** hdb32 file structure:
** 
**   I  M  F  P  R  H
** 
** where:
** 
**   I: identifier section (32 bytes)
**       16 bytes: identifier "hdb32/1.0\0\0\0\0\0\0\0"
**        4 bytes: pid of creator process
**       12 bytes: TAI64N timestamp
** 
**   M: metadata section (8 bytes)
**       4 bytes ntabs: number of hash tables (configured)
**       4 bytes nrecs: total records in file (recorded)
**
**   F: hash function parameter section (8 bytes)
**       4 bytes hf_init: hash function initial value
**       4 bytes hf_mult: hash function multiplier
** 
**   P: hash table pointer section (8 bytes * ntabs)
**       P[0..(ntabs-1)] hash subtable pointers
**       ntabs hash subtable pointers to ntabs subtables
**       each pointer comprised of:
**           4 bytes tbase: offset to base of subtable in section H
**           4 bytes slots: number of slots in subtable
** 
**   R: record data section
**       R[0..(nrecs-1)] records
**       nrecs records, each record comprised of:
**           4 bytes klen: length of key in this record
**           4 bytes dlen: length of data in this record
**           klen bytes: key
**           dlen bytes: data
** 
**   H: hash table section
**       H[0..(ntabs-1)][0..(slots[t])] hash subtables
**       ntabs subtables of slots[t] slots per subtable
**       each slot comprised of:
**           4 bytes: hash of key for record inserted in this slot [*1]
**           4 bytes: rpos, offset to record entry or 0 (empty)
**
**   Notes:
**       1.  The key hash value included in each hash table slot is not
**           strictly required to resolve record searches, but improves
**           search performance in cases of key collisions (enables successive
**           entries in the hash table to be scanned directly, rather than
**           performing a separate record lookup for each collison test).
**
**       2.  Generator pid/timestamp may be used to support hash seed
**           initialization in future implementations.
*/


/*
** --- hdbmk ---
** hdb file generator
*/

/* default parameterization: */

/* default number of hash subtables: */
#define  HDBMK_DEFAULT_NTABS  256

/* default hash function init/mult (fnv32 init/mult parameters): */
/* init = 2166136261: */
#define  HDB_FNV32_BASIS  ((uint32_t)0x0811c9dc5)
#define  HDBMK_DEFAULT_HFINIT  HDB_FNV32_BASIS
/* mult = 16777619: */
#define  HDB_FNV32_PRIME  ((uint32_t)0x01000193)
#define  HDBMK_DEFAULT_HFMULT  HDB_FNV32_PRIME


/*
** struct hdbmk: hdb generator object
** (container for stuff used to fabricate an hdb)
*/
struct hdbmk {
  /* opaque/private internals: */
  struct hdbmk_   *m;
};    
typedef struct hdbmk  hdbmk_t;


/* hdbmk_init()
**   initialize hdbmk_t object H
**     fd: open file descriptor for hdb output
**     ntabs: configure hdb to use ntabs number of hash tables
**     hf_init: configure hash function initial value
**     hf_mult: configure hash function multiplier
**
**   note:
**     if ntabs is 0, a default ntabs is substituted
**
**     if hf_init is 0:
**         if hf_mult is also 0, a default init is substituted
**         else hf_init is 0
**         (hf_init of 0 is accepted iff hf_mult is non-zero)
**
**     if hf_mult is 0, a default multiplier is substituted
**
**   return:
**     0 : success, ioq/fd positioned to begin writing record data
**    -1 : failure, errno set (either malloc() or lseek())
**
*/
extern int hdbmk_init(
  hdbmk_t *H,
  int fd,
  uint32_t ntabs,
  uint32_t hf_init, uint32_t hf_mult);

/* hdbmk_add()
**   add a record to hdb
**     klen: length of key
**     key:  key
**     dlen: length of data
**     data: data
**
**    return:
**      0 : success, no error
**     -1 : failure, errno set
*/
extern int hdbmk_add(
  hdbmk_t *H,
  const uchar_t *key, uint32_t klen,
  const uchar_t *data, uint32_t dlen
);


/* hdbmk_ioqadd()
**   add next record from an ioq reader
**     ioq_in: ioq of input reader positioned for next key/data item
**     klen:   size of key
**     dlen:   size of data
**     sep:    bytes/characters between key/data
**     seplen: size of sep
**
**   return:
**     0 : success, no error
**    -1 : failure, errno set
**
**   notes:
**     "wirespeed" interface for hdbmk
**     no double-copying of data
*/
extern int hdbmk_ioqadd(
  hdbmk_t *H,
  ioq_t     *ioq_in,
  size_t     klen,
  size_t     dlen,
  uchar_t   *sep,
  size_t     seplen);


/* hdbmk_finish()
**   complete generation of hdb after all records added
**
**   return:
**     0 : success
**    -1 : failure, errno set
**
**   notes:
**     * on success, caller should fsync()/close() fd
**     * on return, caller should call hdbmk_clear()
*/
extern int hdbmk_finish(hdbmk_t *H);


/* hdbmk_clear()
**   deallocate/reset an hdbmk_t object
**   notes:
**     * used to free any memory allocated by hdbmk*() functions
**     * should be called after hdbmk_finish() or any error
**     * should be called before reusing an hdbmk_t object with hdbmk_init()
*/
extern void hdbmk_clear(hdbmk_t *H);



#endif /* HDBMK_H */
/* eof (hdbmk.h) */
